/*
Copyright 2015-2016, Manuel Bichler
WWW: <http://dbai.tuwien.ac.at/proj/lpopt/>

This file is part of lpopt.

lpopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

lpopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lpopt.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <type_traits>
#include <string>

#include "asp-domain.hpp"

using namespace std;

namespace domain {

bool var_cmp::operator()(const Variable * lhs, const Variable * rhs) const {
	return (*lhs < *rhs);
}
size_t var_hash::operator() (const Variable &v) const {
	return std::hash<string>()(*(v.name));
}

template<typename T>
vector<T *> *deepVectorCopy(vector<T *>* base) // makes a deep copy of the vector using clone()
{
	static_assert(
			std::is_base_of<AspEntity, T>::value,
			"T must be a descendant of AspEntity"
	);
	if (base == NULL)
		return NULL;
	vector<T *> * result = new vector<T *>();
	result->resize(base->size());
	for (unsigned i = 0; i < result->size(); ++i)
		(*result)[i] = (*base)[i]->clone();
	return result;
}
vector<string *> *deepStringVectorCopy(vector<string *>* base) // makes a deep copy of the vector
{
	if (base == NULL)
		return NULL;
	vector<string *> * result = new vector<string *>();
	result->resize(base->size());
	for (unsigned i = 0; i < result->size(); ++i)
		(*result)[i] = new string(*(*base)[i]);
	return result;
}

template<typename T>
T *clonenull(T* base) // calls clone() on base and returns result. Returns NULL if base is NULL.
		{
	static_assert(
			std::is_base_of<AspEntity, T>::value,
			"T must be a descendant of AspEntity"
	);
	if (base == NULL)
		return NULL;
	return base->clone();
}

Term::~Term() {
}

TermTuple::TermTuple(vector<Term *> *terms) :
		AspEntity() {
	this->terms = terms;
}
TermTuple::TermTuple(TermTuple const &that) :
		terms(deepVectorCopy(that.terms)) {
}
TermTuple::~TermTuple() {
	if (terms != NULL)
	{
		for (vector<Term *>::iterator it = terms->begin(); it != terms->end(); ++it)
			delete *it;
		delete this->terms;
	}
}
TermTuple::TermTuple(TermTuple&& other) :
		TermTuple() {
	swap(*this, other);
}
TermTuple::TermTuple() {
}
TermTuple* TermTuple::clone() const {
	return new TermTuple(*this);
}

TermTuples::TermTuples(vector<TermTuple *> *tuples) :
		AspEntity() {
	this->tuples = tuples;
}
TermTuples::TermTuples(TermTuples const &that) :
		tuples(deepVectorCopy(that.tuples)) {
}
TermTuples::~TermTuples() {
	if (tuples != NULL) {
		for (vector<TermTuple *>::iterator it = tuples->begin();
				it != tuples->end(); ++it)
			delete *it;
		delete this->tuples;
	}
}
TermTuples::TermTuples(TermTuples&& other) :
		TermTuples() {
	swap(*this, other);
}
TermTuples::TermTuples() {
}
TermTuples* TermTuples::clone() const {
	return new TermTuples(*this);
}

UnaryTerm::UnaryTerm(Term *base, UnaryOp type) :
		Term(), base(base), type(type) {
}
UnaryTerm::UnaryTerm(UnaryTerm const &that) :
		base(clonenull(that.base)), type(that.type) {
}
UnaryTerm::~UnaryTerm() {
	delete this->base;
}
UnaryTerm::UnaryTerm(UnaryTerm&& other) {
	swap(*this, other);
}
UnaryTerm* UnaryTerm::clone() const {
	return new UnaryTerm(*this);
}
bool UnaryTerm::bracketingRelevant() const {
	return false;
}

BracketedTerm::BracketedTerm(Term *base) :
		Term() {
	this->base = base;
}
BracketedTerm::BracketedTerm(BracketedTerm const &that) :
		base(clonenull(that.base)) {
}
BracketedTerm::~BracketedTerm() {
	delete this->base;
}
BracketedTerm::BracketedTerm(BracketedTerm&& other) :
		BracketedTerm() {
	swap(*this, other);
}
BracketedTerm::BracketedTerm() {
}
BracketedTerm* BracketedTerm::clone() const {
	return new BracketedTerm(*this);
}
bool BracketedTerm::bracketingRelevant() const {
	return false;
}

BinaryTerm::BinaryTerm(Term *left, string *binopType, Term *right) :
		Term() {
	this->left = left;
	this->right = right;
	this->binopType = binopType;
}
BinaryTerm::BinaryTerm(BinaryTerm const &that) :
		left(clonenull(that.left)), binopType(new string(*(that.binopType))), right(
				clonenull(that.right)) {
}
BinaryTerm::~BinaryTerm() {
	delete this->left;
	delete this->right;
	delete this->binopType;
}
BinaryTerm::BinaryTerm(BinaryTerm&& other) :
		BinaryTerm() {
	swap(*this, other);
}
BinaryTerm::BinaryTerm() {
}
BinaryTerm* BinaryTerm::clone() const {
	return new BinaryTerm(*this);
}
bool BinaryTerm::interval() const {
	return binopType->compare("..") == 0;
}
bool BinaryTerm::arithmetic() const {
	return !interval();
}
bool BinaryTerm::bracketingRelevant() const {
	return true;
}

PoolingTerm::PoolingTerm(vector<Term *> *terms) :
		Term() {
	this->terms = terms;
}
PoolingTerm::PoolingTerm(PoolingTerm const &that) :
		terms(deepVectorCopy(that.terms)) {
}
PoolingTerm::~PoolingTerm() {
	if (terms) {
		for (vector<Term *>::iterator it = terms->begin(); it != terms->end(); ++it)
			delete *it;
		delete this->terms;
	}
}
PoolingTerm::PoolingTerm(PoolingTerm&& other) :
		PoolingTerm() {
	swap(*this, other);
}
PoolingTerm::PoolingTerm() {
}
PoolingTerm* PoolingTerm::clone() const {
	return new PoolingTerm(*this);
}
bool PoolingTerm::bracketingRelevant() const {
	return false;
}

Constant::Constant(string *printable) :
		Term() {
	this->printable = printable;
}
Constant::Constant(Constant const &that) :
		printable(new string(*(that.printable))) {
}
Constant::~Constant() {
	delete this->printable;
}
Constant::Constant(Constant&& other) :
		Constant() {
	swap(*this, other);
}
Constant::Constant() {
}
Constant* Constant::clone() const {
	return new Constant(*this);
}
bool Constant::bracketingRelevant() const {
	return false;
}

Variable::Variable(string *name) :
		Term() {
	this->name = name;
}
Variable::Variable(Variable const &that) :
		name(new string(*(that.name))) {
}
Variable::~Variable() {
	delete this->name;
}
Variable::Variable(Variable&& other) :
		Variable() {
	swap(*this, other);
}
Variable::Variable() {
}
Variable* Variable::clone() const {
	return new Variable(*this);
}
bool Variable::bracketingRelevant() const {
	return false;
}
const string& Variable::getName() const {
	return *name;
}

bool operator==(const Variable& lhs, const Variable& rhs) {
	return *(lhs.name) == *(rhs.name);
}
bool operator!=(const Variable& lhs, const Variable& rhs) {
	return !operator==(lhs, rhs);
}
bool operator<(const Variable& lhs, const Variable& rhs) {
	return *(lhs.name) < *(rhs.name);
}
bool operator>(const Variable& lhs, const Variable& rhs) {
	return operator<(rhs, lhs);
}
bool operator<=(const Variable& lhs, const Variable& rhs) {
	return !operator>(lhs, rhs);
}
bool operator>=(const Variable& lhs, const Variable& rhs) {
	return !operator<(lhs, rhs);
}

AnonymousVariable::AnonymousVariable() :
		Term() {
}
AnonymousVariable::AnonymousVariable(AnonymousVariable const &/*that*/) :
		Term() {
}
AnonymousVariable::~AnonymousVariable() {
}
AnonymousVariable::AnonymousVariable(AnonymousVariable&& other) :
		AnonymousVariable() {
	swap(*this, other);
}
AnonymousVariable* AnonymousVariable::clone() const {
	return new AnonymousVariable(*this);
}
bool AnonymousVariable::bracketingRelevant() const {
	return false;
}

inline bool operator==(const AnonymousVariable& /*lhs*/,
		const AnonymousVariable& /*rhs*/) {
	return true;
}
inline bool operator!=(const AnonymousVariable& lhs,
		const AnonymousVariable& rhs) {
	return !operator==(lhs, rhs);
}
inline bool operator<(const AnonymousVariable& /*lhs*/,
		const AnonymousVariable& /*rhs*/) {
	return false;
}
inline bool operator>(const AnonymousVariable& lhs,
		const AnonymousVariable& rhs) {
	return operator<(rhs, lhs);
}
inline bool operator<=(const AnonymousVariable& lhs,
		const AnonymousVariable& rhs) {
	return !operator>(lhs, rhs);
}
inline bool operator>=(const AnonymousVariable& lhs,
		const AnonymousVariable& rhs) {
	return !operator<(lhs, rhs);
}

Function::Function(string *name, TermTuples *tuples) :
		Term() {
	this->name = name;
	this->tuples = tuples;
}
Function::Function(Function const &that) :
		name(new string(*(that.name))), tuples(clonenull(that.tuples)) {
}
Function::~Function() {
	delete this->name;
	delete this->tuples;
}
Function::Function(Function&& other) :
		Function() {
	swap(*this, other);
}
Function::Function() {
}
Function* Function::clone() const {
	return new Function(*this);
}
bool Function::bracketingRelevant() const {
	return false;
}

BodyElement::~BodyElement() {
}
varset* BodyElement::variables() const
{
  return new varset(publicVariables);
}
const varset& BodyElement::variablesFast() const
{
  return publicVariables;
}

NafLiteral::NafLiteral() : BodyElement(), safetyCh(NULL) {
}

ClassicalLiteral::ClassicalLiteral(string *name, bool strongNegated,
		bool nafNegated, TermTuples *tuples) :
		NafLiteral() {
	this->name = name;
	this->strongNegated = strongNegated, this->nafNegated = nafNegated, this->tuples =
			tuples;
	updatePublicVariables();
}
ClassicalLiteral::ClassicalLiteral(string *name, bool strongNegated,
		bool nafNegated) :
		ClassicalLiteral(name, strongNegated, nafNegated, NULL) {
}
ClassicalLiteral::ClassicalLiteral(ClassicalLiteral const &that) : NafLiteral(that), 
		name(new string(*(that.name))), strongNegated(that.strongNegated), nafNegated(
				that.nafNegated), tuples(clonenull(that.tuples)) {
	updatePublicVariables();
}
ClassicalLiteral::~ClassicalLiteral() {
	delete this->name;
	delete this->tuples;
}
ClassicalLiteral::ClassicalLiteral(ClassicalLiteral&& other) :
		ClassicalLiteral() {
	swap(*this, other);
}
ClassicalLiteral::ClassicalLiteral() : NafLiteral() {
}
ClassicalLiteral* ClassicalLiteral::clone() const {
	return new ClassicalLiteral(*this);
}

Statement::~Statement() {
}

Program::Program(vector<Statement *> *statements, vector<string *> *metastatements, ClassicalLiteral *query) :
		AspEntity(), statements(statements), metastatements(metastatements), query(query) {
}
Program::Program(vector<Statement *> *statements, ClassicalLiteral *query) :
		Program(statements, new vector<string*>(), query) {
}
Program::Program() :
		Program(new vector<Statement *>(), new vector<string*>(), NULL) {
}
Program::Program(vector<Statement *> *statements, vector<string *> *metastatements) :
		Program(statements, metastatements, NULL) {
}
Program::Program(vector<Statement *> *statements) :
		Program(statements, new vector<string*>(), NULL) {
}
Program::Program(ClassicalLiteral *query) :
		Program(new vector<Statement *>(), query) {
}

Program::Program(Program const &that) :
		statements(deepVectorCopy(that.statements)), metastatements(deepStringVectorCopy(that.metastatements)), query(clonenull(that.query)) {
}
Program::~Program() {
	if (metastatements) {
		for (vector<string *>::iterator it = metastatements->begin();
				it != metastatements->end(); ++it)
			delete *it;
		delete this->metastatements;
	}
	if(statements) {
		for (vector<Statement *>::iterator it = statements->begin();
				it != statements->end(); ++it)
			delete *it;
		delete this->statements;
	}
	delete this->query;
}
Program::Program(Program&& other) :
		Program() {
	swap(*this, other);
}
Program* Program::clone() const {
	return new Program(*this);
}

void Program::replaceStatements(vector<Statement *> *newStatements) const{
	*statements = *newStatements;
}
const Statement& Program::getStatement(int n) const {
	return *(*statements)[n];
}
vector<Statement *> Program::getStatements(){
	return *statements;
}
unsigned int Program::statementsSize() const {
	return statements->size();
}
Program* Program::cloneQueryAndMetastatements(vector<Statement*>* stmts) const
{
	return new Program(stmts, deepStringVectorCopy(metastatements), clonenull(query));
}

WeightAtLevel::WeightAtLevel(Term *weight, Term *level, TermTuple *tuple) :
		AspEntity() {
	this->weight = weight;
	this->level = level;
	this->tuple = tuple;
}
WeightAtLevel::WeightAtLevel(Term *weight) :
		WeightAtLevel(weight, NULL, NULL) {
}
WeightAtLevel::WeightAtLevel(Term *weight, Term *level) :
		WeightAtLevel(weight, level, NULL) {
}
WeightAtLevel::WeightAtLevel(Term *weight, TermTuple *tuple) :
		WeightAtLevel(weight, NULL, tuple) {
}
WeightAtLevel::WeightAtLevel(WeightAtLevel const &that) :
		weight(clonenull(that.weight)), level(clonenull(that.level)), tuple(
				clonenull(that.tuple)) {
}
WeightAtLevel::~WeightAtLevel() {
	delete this->weight;
	delete this->level;
	delete this->tuple;
}
WeightAtLevel::WeightAtLevel(WeightAtLevel&& other) :
		WeightAtLevel() {
	swap(*this, other);
}
WeightAtLevel::WeightAtLevel() {
}
WeightAtLevel* WeightAtLevel::clone() const {
	return new WeightAtLevel(*this);
}

LiteralBody::LiteralBody(vector<NafLiteral *> *literals) :
		AspEntity(), literals(literals) {
	calcVariables();
}
LiteralBody::LiteralBody() :
		AspEntity(), globalVariables() {
	this->literals = new vector<NafLiteral*>();
}
LiteralBody::LiteralBody(LiteralBody const &that) :
		literals(deepVectorCopy(that.literals)) {
	calcVariables();
}
LiteralBody::~LiteralBody() {
	if (literals == NULL) return;
	for (vector<NafLiteral *>::iterator it = literals->begin();
			it != literals->end(); ++it)
		delete *it;
	delete this->literals;
}
LiteralBody::LiteralBody(LiteralBody&& other) :
		LiteralBody() {
	swap(*this, other);
}
LiteralBody::LiteralBody(Body* from) : LiteralBody() {
	while(!from->elements->empty())
	{
		literals->push_back((NafLiteral *)(from->elements->front()));
		from->elements->erase(from->elements->begin());
	}
	calcVariables();
	delete from;
}
LiteralBody* LiteralBody::clone() const {
	return new LiteralBody(*this);
}

bool LiteralBody::empty() const {
	return literals->empty();
}

int LiteralBody::size() const
{
	return literals->size();
}
Body* LiteralBody::toRegularBody()
{
	Body *b = new Body(new vector<BodyElement*>(literals->begin(), literals->end()));
	delete literals;
	literals = NULL;
	return b;
}

void LiteralBody::insert(NafLiteral* literal)
{
	literals->push_back(literal);
	varset* newVars = ((BodyElement*)literal)->variables();
	globalVariables.insert(newVars->begin(), newVars->end());
	delete newVars;
}

void LiteralBody::combine(LiteralBody* other)
{
	literals->insert(literals->end(), other->literals->begin(), other->literals->end());
	calcVariables();
	delete other->literals;
	other->literals = NULL;
	delete other;
} // moves the other's literals into this, deleting other.

OptimizeElement::OptimizeElement(WeightAtLevel *weightAtLevel,
		LiteralBody *literals) :
		AspEntity() {
	this->weightAtLevel = weightAtLevel;
	this->literals = literals;
}
OptimizeElement::OptimizeElement(OptimizeElement const &that) :
		weightAtLevel(clonenull(that.weightAtLevel)), literals(
				clonenull(that.literals)) {
}
OptimizeElement::~OptimizeElement() {
	delete this->weightAtLevel;
	delete this->literals;
}
OptimizeElement::OptimizeElement(OptimizeElement&& other) :
		OptimizeElement() {
	swap(*this, other);
}
OptimizeElement::OptimizeElement() {
}
OptimizeElement* OptimizeElement::clone() const {
	return new OptimizeElement(*this);
}
void OptimizeElement::addToBody(NafLiteral *l)
{
	literals->insert(l);
}
OptimizeElement* OptimizeElement::cloneHead(Body* body) const
{
	return new OptimizeElement(weightAtLevel->clone(), new LiteralBody(body));
}

OptimizeStatement::OptimizeStatement(string *optimizeType,
		vector<OptimizeElement *> *elements) :
		Statement() {
	this->optimizeType = optimizeType;
	this->elements = elements;
}
OptimizeStatement::OptimizeStatement(OptimizeStatement const &that) :
		optimizeType(new string(*(that.optimizeType))), elements(
				deepVectorCopy(that.elements)) {
}
OptimizeStatement::~OptimizeStatement() {
	if (elements) {
		for (vector<OptimizeElement *>::iterator it = elements->begin();
				it != elements->end(); ++it)
			delete *it;
	}
	delete this->optimizeType;
	delete this->elements;
}
OptimizeStatement::OptimizeStatement(OptimizeStatement&& other) :
		OptimizeStatement() {
	swap(*this, other);
}
OptimizeStatement::OptimizeStatement() {
}
OptimizeStatement* OptimizeStatement::clone() const {
	return new OptimizeStatement(*this);
}
void OptimizeStatement::addToBody(NafLiteral *l)
{
	if (elements->size() != 1)
		return;  // TODO warning/error
	(*elements)[0]->addToBody(l);
}
Statement* OptimizeStatement::cloneHead(Body* body) const
{
	if (elements->size() != 1)
		return NULL; // TODO warning/error
	return new OptimizeStatement(new string(*optimizeType), new vector<OptimizeElement *>(1, (*elements)[0]->cloneHead(body)));
}

Body::Body(vector<BodyElement *> *elements) :
		AspEntity(), elements(elements) {
	calcVariables();
}
Body::Body(Body const &that) :
		elements(deepVectorCopy(that.elements)) {
	calcVariables();
}
Body::~Body() {
	if (elements) {
		for (vector<BodyElement *>::iterator it = elements->begin();
				it != elements->end(); ++it)
			delete *it;
		delete this->elements;
	}
}
Body::Body(Body&& other) :
		Body() {
	swap(*this, other);
}
Body::Body() : Body(new vector<BodyElement *>()) {
}
Body* Body::clone() const {
	return new Body(*this);
}
bool Body::empty() const {
	return elements->empty();
}
void Body::add(BodyElement *el)
{
	elements->push_back(el);
	varset* newVars = el->variables();
	globalVariables.insert(newVars->begin(), newVars->end());
	delete newVars;
}


WeakConstraint::WeakConstraint(Body *body, WeightAtLevel *weightAtLevel) :
		Statement() {
	this->body = body;
	this->weightAtLevel = weightAtLevel;
}
WeakConstraint::WeakConstraint(WeakConstraint const &that) :
		body(clonenull(that.body)), weightAtLevel(clonenull(that.weightAtLevel)) {
}
WeakConstraint::~WeakConstraint() {
	delete this->body;
	delete this->weightAtLevel;
}
WeakConstraint::WeakConstraint(WeakConstraint&& other) :
		WeakConstraint() {
	swap(*this, other);
}
WeakConstraint::WeakConstraint() {
}
WeakConstraint* WeakConstraint::clone() const {
	return new WeakConstraint(*this);
}
void WeakConstraint::addToBody(NafLiteral *l)
{
	body->add(l);
}
Statement* WeakConstraint::cloneHead(Body* body) const
{
	return new WeakConstraint(body, weightAtLevel->clone());
}

RuleHead::~RuleHead() {
}


Rule::Rule(RuleHead *head, Body *body) :
		Statement() {
	this->head = head;
	this->body = body;
}

Rule::Rule(Rule const &that) :
		head(clonenull(that.head)), body(clonenull(that.body)) {
}

Rule::~Rule() {
	delete this->head;
	delete this->body;
}

Rule::Rule(Rule&& other) :
		Rule() {
	swap(*this, other);
}

Rule::Rule() {
}

Rule* Rule::clone() const {
	return new Rule(*this);
}

void Rule::addToBody(NafLiteral *l)
{
	body->add(l);
}
Statement* Rule::cloneHead(Body* body) const
{
	return new Rule(head->clone(), body);
}

Disjunction::Disjunction(vector<ClassicalLiteral *> *literals) :
		RuleHead() {
	this->literals = literals;
}
Disjunction::Disjunction(Disjunction const &that) :
		literals(deepVectorCopy(that.literals)) {
}
Disjunction::~Disjunction() {
	if (literals) {
		for (vector<ClassicalLiteral *>::iterator it = literals->begin();
				it != literals->end(); ++it)
			delete *it;
		delete this->literals;
	}
}
Disjunction::Disjunction(Disjunction&& other) :
		Disjunction() {
	swap(*this, other);
}
Disjunction::Disjunction() {
}
Disjunction* Disjunction::clone() const {
	return new Disjunction(*this);
}
bool Disjunction::empty() {
	return literals->empty();
}

Guard::Guard(string *binop, Term *term, bool isLeft) :
		AspEntity() {
	this->binop = binop;
	this->term = term;
	this->left = isLeft;
}
Guard::Guard(Guard const &that) :
		binop(new string(*(that.binop))), term(clonenull(that.term)), left(
				that.left) {
}
Guard::~Guard() {
	delete this->binop;
	delete this->term;
}
Guard::Guard(Guard&& other) :
		Guard() {
	swap(*this, other);
}
Guard::Guard() {
}
bool Guard::isLeft() const {
	return this->left;
}
Guard* Guard::clone() const {
	return new Guard(*this);
}
bool Guard::equal() const {
	return *binop == "=";
}

ChoiceElement::ChoiceElement(ClassicalLiteral *literal,
		LiteralBody *dependencyLiterals) :
		AspEntity() {
	this->literal = literal;
	this->dependencyLiterals = dependencyLiterals;
}
ChoiceElement::ChoiceElement(ChoiceElement const &that) :
		literal(clonenull(that.literal)), dependencyLiterals(
				clonenull(that.dependencyLiterals)) {
}
ChoiceElement::~ChoiceElement() {
	delete this->literal;
	delete this->dependencyLiterals;
}
ChoiceElement::ChoiceElement(ChoiceElement&& other) :
		ChoiceElement() {
	swap(*this, other);
}
ChoiceElement::ChoiceElement() {
}
ChoiceElement* ChoiceElement::clone() const {
	return new ChoiceElement(*this);
}
Choice::Choice(vector<ChoiceElement *> *elements, Guard *guard_l,
		Guard *guard_r) :
		RuleHead() {
	this->elements = elements;
	this->guard_l = guard_l;
	this->guard_r = guard_r;
}
Choice::Choice(vector<ChoiceElement *> *elements) :
		Choice(elements, NULL, NULL) {
}
Choice::Choice(vector<ChoiceElement *> *elements, Guard *guard) :
		Choice(elements) {
	if (guard->isLeft())
		this->guard_l = guard;
	else
		this->guard_r = guard;
}
Choice::Choice(Choice const &that) :
		elements(deepVectorCopy(that.elements)), guard_l(
				clonenull(that.guard_l)), guard_r(clonenull(that.guard_r)) {
}
Choice::~Choice() {
	if (elements) {
		for (vector<ChoiceElement *>::iterator it = elements->begin();
				it != elements->end(); ++it)
			delete *it;
		delete this->elements;
	}
	delete this->guard_l;
	delete this->guard_r;
}
Choice::Choice(Choice&& other) :
		Choice() {
	swap(*this, other);
}
Choice::Choice() {
}
Choice* Choice::clone() const {
	return new Choice(*this);
}
bool Choice::empty() {
	return false; // a choice head consists at least of {}
}
AggregateElement::AggregateElement(TermTuple *termTuple,
		LiteralBody *dependencyLiterals) :
		AspEntity() {
	this->termTuple = termTuple;
	this->dependencyLiterals = dependencyLiterals;
}
AggregateElement::AggregateElement(TermTuple *termTuple) :
		AggregateElement(termTuple, new LiteralBody()) {
}
AggregateElement::AggregateElement(LiteralBody *dependencyLiterals) :
		AggregateElement(NULL, dependencyLiterals) {
}
AggregateElement::AggregateElement(AggregateElement const &that) :
		termTuple(clonenull(that.termTuple)), dependencyLiterals(
				clonenull(that.dependencyLiterals)) {
}
AggregateElement::~AggregateElement() {
	delete this->termTuple;
	delete this->dependencyLiterals;
}
AggregateElement::AggregateElement(AggregateElement&& other) :
		AggregateElement() {
	swap(*this, other);
}
AggregateElement::AggregateElement() {
}
AggregateElement* AggregateElement::clone() const {
	return new AggregateElement(*this);
}
Aggregate::Aggregate(string *function, bool nafNegated,
		vector<AggregateElement *> *elements, Guard *guard_l, Guard *guard_r) :
		BodyElement() {
	this->function = function;
	this->nafNegated = nafNegated;
	this->elements = elements;
	this->guard_l = guard_l;
	this->guard_r = guard_r;
	updatePublicVariables();
}
Aggregate::Aggregate(string *function, bool nafNegated,
		vector<AggregateElement *> *elements, Guard *guard) :
		Aggregate(function, nafNegated, elements) {
	if (guard->isLeft())
		this->guard_l = guard;
	else
		this->guard_r = guard;
	updatePublicVariables();
}
Aggregate::Aggregate(string *function, bool nafNegated,
		vector<AggregateElement *> *elements) :
		Aggregate(function, nafNegated, elements, NULL, NULL) {
}
Aggregate::Aggregate(Aggregate const &that) : BodyElement(that),
		function(new string(*(that.function))), nafNegated(that.nafNegated), elements(
				deepVectorCopy(that.elements)), guard_l(
				clonenull(that.guard_l)), guard_r(clonenull(that.guard_r)) {
	updatePublicVariables();
}
Aggregate::~Aggregate() {
	if (elements) {
		for (vector<AggregateElement *>::iterator it = elements->begin();
				it != elements->end(); ++it)
			delete *it;
	}
	delete this->function;
	delete this->elements;
}
Aggregate::Aggregate(Aggregate&& other) :
		Aggregate() {
	swap(*this, other);
}
Aggregate::Aggregate() : BodyElement() {
}
Aggregate* Aggregate::clone() const {
	return new Aggregate(*this);
}

BuiltinAtom::BuiltinAtom(Term *t1, string *binop, Term *t2) :
		NafLiteral() {
	this->t1 = t1;
	this->binop = binop;
	this->t2 = t2;
	updatePublicVariables();
}
BuiltinAtom::BuiltinAtom(BuiltinAtom const &that) : NafLiteral(that) {
  t1 = clonenull(that.t1);
  t2 = clonenull(that.t2);
  binop = new string(*(that.binop));
  updatePublicVariables();
  
}
BuiltinAtom::~BuiltinAtom() {
	delete this->t1;
	delete this->t2;
	delete this->binop;
}
BuiltinAtom::BuiltinAtom(BuiltinAtom&& other) :
		BuiltinAtom() {
	swap(*this, other);
}
BuiltinAtom::BuiltinAtom() : NafLiteral() {
}
BuiltinAtom* BuiltinAtom::clone() const {
	return new BuiltinAtom(*this);
}
bool BuiltinAtom::equal() const {
	return *binop == "=";
}

} // domain

