/*
Copyright 2015-2016, Manuel Bichler
WWW: <http://dbai.tuwien.ac.at/proj/lpopt/>

This file is part of lpopt.

lpopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

lpopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lpopt.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ASP_DOMAIN_H_
#define ASP_DOMAIN_H_

#include <set>
#include <ostream>
#include <vector>

//TODO: public getters

using namespace std;

namespace parser
{
  class Parser; // friend the parser in our classes so it can set private member elements
}

namespace domain
{
  class SafetyCharacterization;
  class StringGenerator;

  class AspEntity;
  class Term;
  class UnaryTerm;
  class BracketedTerm;
  class BinaryTerm;
  class PoolingTerm;
  class Constant;
  class Variable;
  class AnonymousVariable;
  class Function;
  class TermTuples;
  class TermTuple;
  class Program;
  class Statement; 
  class OptimizeStatement;
  class WeakConstraint;
  class Rule;
  class Body;
  class BodyElement;
  class OptimizeElement;
  class WeightAtLevel;
  class RuleHead;
  class Disjunction;
  class NafLiteral;
  class Choice;
  class ClassicalLiteral;
  class Aggregate;
  class Guard;
  class ChoiceElement;
  class AggregateElement;
  class BuiltinAtom;
  class LiteralBody;
  
  enum UnaryOp { NEGATION, ABSOLUTE, COMPLEMENT }; // for UnaryTerm
  
  struct var_cmp // functor comparing two Variable pointers for <
  {
    bool operator() (const Variable * lhs, const Variable * rhs) const;    
  };
  struct var_hash // functor generating hash of Variable pointer for use in a hash set
  {
    size_t operator() (const Variable &v) const;    
  };
  template <typename T>
  struct pair_cmp // functor comparing two pair<T *, int> by their int-parts
  {
    bool operator() (pair<T *, int> lhs, pair<T *, int> rhs) const;    
  };

  typedef set<const Variable *, var_cmp> varset;

  template <typename T>
  T *clonenull(T* base); // calls clone() on base and returns result. Returns NULL is base is NULL.
  
  bool empty_intersection(const varset& x, const varset& y); // returns true if no intersection, false otherwise.

  class AspEntity // abstract superclass of all ASP entities
  {
  public:
    virtual void write(ostream &stream) const = 0; // writes this entity in ASP syntax to stream. recursively called.
    virtual AspEntity* clone() const = 0; // deep copy
  };

  class Term : public AspEntity {
    friend class parser::Parser;
  public:
    virtual ~Term();
    virtual varset* variables() const = 0; // all variables in this term
    virtual Term* clone() const = 0; // deep copy
    virtual vector<Term*>* splitPooling() = 0; // if this term has no pooling (deep search), does nothing and returns an empty vector. If it contains pooling, splits all pooling occurrences in this term into multiple terms. this is changed to one result and the other results are returned.
    virtual bool bracketingRelevant() const = 0; // returns false if and only if the fact holds that when this term is surrounded by brackets, omitting these bracktes would not change semantics in any possible case. Otherwise, returns true. Practically, only BinaryTerm returns true here.
    virtual SafetyCharacterization* safety() const = 0; // returns a characterization of this term (subterms combined). Safety characterization of terms include no safety rules, i.e. all variables are either safe or unsafe.
  }; // abstract class
    
  class TermTuple : public AspEntity
  {
    friend class parser::Parser;
  public:
    TermTuple(vector<Term *> *terms);
    TermTuple(TermTuple const &that);
    virtual ~TermTuple();
    TermTuple& operator=(TermTuple that);
    TermTuple(TermTuple&& other);
  private:
    TermTuple();
  public:
    friend void swap(TermTuple& first, TermTuple& second);
    void write(ostream &stream) const;
    varset* variables() const;
    TermTuple* clone() const;
    vector<TermTuple*>* splitPooling();
    SafetyCharacterization* safety() const;
  private:
    vector<Term *> *terms;
  };
  
  
  class TermTuples : public AspEntity
  {
    friend class parser::Parser;
  public:
    TermTuples(vector<TermTuple *> *tuples);
    TermTuples(TermTuples const &that);
    virtual ~TermTuples();
    TermTuples& operator=(TermTuples that);
    TermTuples(TermTuples&& other);
  private:
    TermTuples();
  public:
    friend void swap(TermTuples& first, TermTuples& second);
    void write(ostream &stream) const;
    varset* variables() const;
    TermTuples* clone() const;
    vector<TermTuples*>* splitPooling();
    SafetyCharacterization* safety() const;
  private:
    vector<TermTuple *> *tuples;
  };
  
  class UnaryTerm : public Term
  {
    friend class parser::Parser;
  public:
    UnaryTerm(Term *base, UnaryOp type);
    UnaryTerm(UnaryTerm const &that);
    virtual ~UnaryTerm();
    UnaryTerm& operator=(UnaryTerm that);
    UnaryTerm(UnaryTerm&& other);
  private:
//     UnaryTerm();
  public:
    friend void swap(UnaryTerm& first, UnaryTerm& second);
    void write(ostream &stream) const;
    varset* variables() const;
    UnaryTerm* clone() const;
    vector<Term*>* splitPooling();
    bool bracketingRelevant() const;
    SafetyCharacterization* safety() const; // safe for '-' (unary minus), unsafe for || (absolute value), ~ (bitwise complement)
  private:
    Term *base;
    UnaryOp type;
  };
  
  class BracketedTerm : public Term
  {
    friend class parser::Parser;
  public:
    BracketedTerm(Term *base);
    BracketedTerm(BracketedTerm const &that);
    virtual ~BracketedTerm();
    BracketedTerm& operator=(BracketedTerm that);
    BracketedTerm(BracketedTerm&& other);
  private:
    BracketedTerm();
  public:
    friend void swap(BracketedTerm& first, BracketedTerm& second);
    void write(ostream &stream) const;
    varset* variables() const;
    BracketedTerm* clone() const;
    vector<Term*>* splitPooling();
    bool bracketingRelevant() const;
    SafetyCharacterization* safety() const;
  private:
    Term *base;
  };
  
  class BinaryTerm : public Term
  {
    friend class parser::Parser;
  public:
    BinaryTerm(Term *left, string *binopType, Term *right);
    BinaryTerm(BinaryTerm const &that);
    virtual ~BinaryTerm();
    BinaryTerm& operator=(BinaryTerm that);
    BinaryTerm(BinaryTerm&& other);
  private:
    BinaryTerm();
  public:
    friend void swap(BinaryTerm& first, BinaryTerm& second);
    void write(ostream &stream) const;
    varset* variables() const;
    BinaryTerm* clone() const;
    bool interval() const;
    bool arithmetic() const;
    vector<Term*>* splitPooling();
    bool bracketingRelevant() const;
    SafetyCharacterization* safety() const;
  private:
    Term *left;
    string *binopType;
    Term *right;
  };
  
  class PoolingTerm : public Term
  {
    friend class parser::Parser;
  public:
    PoolingTerm(vector<Term *> *terms);
    PoolingTerm(PoolingTerm const &that);
    virtual ~PoolingTerm();
    PoolingTerm& operator=(PoolingTerm that);
    PoolingTerm(PoolingTerm&& other);
  private:
    PoolingTerm();
  public:
    friend void swap(PoolingTerm& first, PoolingTerm& second);
    PoolingTerm* clone() const;
    void write(ostream &stream) const;
    varset* variables() const;
    vector<Term*>* splitPooling();
    bool bracketingRelevant() const;
    SafetyCharacterization* safety() const;
  private:
    vector<Term *> *terms;
  };
  
  class Constant : public Term // like 4, "teststring" or mama
  {
    friend class parser::Parser;
  public:
    Constant(string *printable);
    Constant(Constant const &that);
    virtual ~Constant();
    Constant& operator=(Constant that);
    Constant(Constant&& other);
  private:
    Constant();
  public:
    friend void swap(Constant& first, Constant& second);
    Constant* clone() const;
    void write(ostream &stream) const;
    varset* variables() const; // empty
    vector<Term*>* splitPooling();
    bool bracketingRelevant() const;
    SafetyCharacterization* safety() const;
  private:
    string *printable;
  };
  
  class Variable : public Term
  {
    friend class parser::Parser;
    friend class var_hash;
  public:
    Variable(string *name);
    Variable(Variable const &that);
    virtual ~Variable();
    Variable& operator=(Variable that);
    Variable(Variable&& other);
  private:
    Variable();
  public:
    friend void swap(Variable& first, Variable& second);
    friend bool operator==(const Variable& lhs, const Variable& rhs);
    friend bool operator<(const Variable& lhs, const Variable& rhs);
    void write(ostream &stream) const;
    varset* variables() const; // only itself contained
    Variable* clone() const;
    vector<Term*>* splitPooling();
    bool bracketingRelevant() const;
    SafetyCharacterization* safety() const;
    const string& getName() const;
  private:
    string *name;
  };
  bool operator==(const Variable& lhs, const Variable& rhs);
  bool operator!=(const Variable& lhs, const Variable& rhs);
  bool operator< (const Variable& lhs, const Variable& rhs);
  bool operator> (const Variable& lhs, const Variable& rhs);
  bool operator<=(const Variable& lhs, const Variable& rhs);
  bool operator>=(const Variable& lhs, const Variable& rhs);

  class AnonymousVariable : public Term
  {
    friend class parser::Parser;
  public:
    AnonymousVariable();
    AnonymousVariable(AnonymousVariable const &that);
    virtual ~AnonymousVariable();
    AnonymousVariable& operator=(AnonymousVariable that);
    AnonymousVariable(AnonymousVariable&& other);
  public:
    friend void swap(AnonymousVariable& first, AnonymousVariable& second);
    void write(ostream &stream) const;
    varset* variables() const; // empty
    AnonymousVariable* clone() const;
    vector<Term*>* splitPooling();
    bool bracketingRelevant() const;
    SafetyCharacterization* safety() const;
  };
  
  inline bool operator==(const AnonymousVariable& lhs, const AnonymousVariable& rhs);
  inline bool operator!=(const AnonymousVariable& lhs, const AnonymousVariable& rhs);
  inline bool operator< (const AnonymousVariable& lhs, const AnonymousVariable& rhs);
  inline bool operator> (const AnonymousVariable& lhs, const AnonymousVariable& rhs);
  inline bool operator<=(const AnonymousVariable& lhs, const AnonymousVariable& rhs);
  inline bool operator>=(const AnonymousVariable& lhs, const AnonymousVariable& rhs);
  
  class Function : public Term
  {
    friend class parser::Parser;
  public:
    Function(string *name, TermTuples *tuples);
    Function(Function const &that);
    virtual ~Function();
    Function& operator=(Function that);
    Function(Function&& other);
  private:
    Function();
  public:
    friend void swap(Function& first, Function& second);
    void write(ostream &stream) const;
    varset* variables() const;
    Function* clone() const;
    vector<Term*>* splitPooling();
    bool bracketingRelevant() const;
    SafetyCharacterization* safety() const;
  private:
    string *name;
    TermTuples *tuples;
  };
  
  class BodyElement : public AspEntity
  {
    friend class parser::Parser;
  public:
    virtual ~BodyElement();
    varset* variables() const; // public variables in this body element. might not be all, but at least all safe are inside. The returned object is a new copy.
    const varset& variablesFast() const; // same as variables(), but for performance reasons (no new object creation).
    virtual varset* variables(const varset& globalVars) const = 0; // all publically visible variables in this body element
    virtual bool globalVarsNotEmptyAndIncludedIn(const varset& globalVariables, const varset& s) const = 0; // checks if the variables of variables(globalVars) are not empty and fully included in s
    virtual BodyElement* clone() const = 0; // deep copy
    friend void swap(BodyElement& first, BodyElement& second); // subclass-swaps must call this
    virtual vector<BodyElement*>* splitPooling() = 0; // splits this body element into multiple body elements if it contains pooling. this is changed to one result, the others are returned. If no pooling occurs, changes nothing and returns an empty vector.
    virtual const SafetyCharacterization* safety(const varset& globalVars) = 0; // returns the safety characterization of this element. Implementations depending on globalVars are likely to create a new object on each call. Others may lazy-generate a single object for all the object lieftime. Do not delete the returned object. After use, call dispose().
    virtual void dispose(const SafetyCharacterization* safety) const = 0; // takes care of properly disposing of the SafetyCharacterization. Subclasses that generate a new object on each call of safety() may delete it; others may not.
    virtual vector<Rule*>* optimizeInnerDomains(StringGenerator& gen, const varset& globalVars) = 0; // if this is not an aggregate or an non-optimizable aggregate, does nothing (returns NULL). If an optimizable aggregate, changes it (introduces new predicates) and returns the induction rules for the new predicates.
  protected:
    varset publicVariables; // public variables in this body element. might not be all, but at least all safe are inside. Implementers should take care to have publicVariables conform at all times.
  }; // abstract class

  class NafLiteral : public BodyElement
  {
    friend class parser::Parser;
  public:
    NafLiteral(); // subclasses must use this constructor
    virtual ~NafLiteral();
    NafLiteral(NafLiteral const &that); // copy constructor. subclasses must use this in their copy constructors.
    virtual varset* variables(const varset& globalVars) const; // all publically visible variables in this body element (same as variables())
    virtual bool globalVarsNotEmptyAndIncludedIn(const varset& globalVariables, const varset& s) const; // checks if the variables of variables(globalVars) are not empty and fully included in s
    virtual NafLiteral* clone() const = 0; // deep copy
    friend void swap(NafLiteral& first, NafLiteral& second); // subclass-swaps must call this
    virtual vector<BodyElement*>* splitPooling() = 0;
    virtual vector<NafLiteral*>* splitPoolingNaf() = 0;
    virtual vector<Rule*>* optimizeInnerDomains(StringGenerator& gen, const varset& globalVars); // does nothing (returns NULL).
    const SafetyCharacterization* safety(const varset& globalVars); // lazy-initializes safetyCh and returns same object each call
    void dispose(const SafetyCharacterization* safety) const; // does nothing
  protected:
    virtual const SafetyCharacterization* computeSafetyCharacterization() const = 0; // subclasses implement this. Each call, they compute a new characterization object and return it.
  private:
    const SafetyCharacterization* safetyCh; // lifetime-unchanged
  }; // abstract class

  class ClassicalLiteral : public NafLiteral
  {
    friend class parser::Parser;
  public:
    ClassicalLiteral(string *name, bool strongNegated, bool nafNegated, TermTuples *tuples);
    ClassicalLiteral(string *name, bool strongNegated, bool nafNegated);
    ClassicalLiteral(ClassicalLiteral const &that);
    virtual ~ClassicalLiteral();
    ClassicalLiteral& operator=(ClassicalLiteral that);
    ClassicalLiteral(ClassicalLiteral&& other);
  private:
    ClassicalLiteral();
  public:
    friend void swap(ClassicalLiteral& first, ClassicalLiteral& second);
    void write(ostream &stream) const;
    ClassicalLiteral* clone() const;
    template <typename T> vector<T *> *splitPooling(); // T must be supertype of ClassicalLiteral
    vector<BodyElement*>* splitPooling();
    vector<NafLiteral*>* splitPoolingNaf();
  protected:
    const SafetyCharacterization* computeSafetyCharacterization() const;
  private:
    void updatePublicVariables(); // sets publicVariables
  private:
    string *name;
    bool strongNegated, nafNegated;
    TermTuples *tuples; // may be NULL
  };

  class Statement : public AspEntity
  {
    friend class parser::Parser;
  public:
    virtual ~Statement();
    virtual varset* variables() const = 0; // all publically visible variables in this statement
    virtual Statement* clone() const = 0; // deep copy
    virtual vector<Statement*>* splitPooling() = 0; // splits this statement into multiple statements if it contains pooling. this is changed to one result, the others are returned. If no pooling occurs, changes nothing and returns an empty vector.
    virtual SafetyCharacterization* safety() const = 0; // returns the safety characterization of this statement. splitPooling() must be performed prior to calling this function. only the body is used to deduce the safety characterization. (within weak constraints and optimize statements, the weight@level is considered the head)
    virtual vector<Rule*>* optimizeInnerDomains(StringGenerator& gen) = 0; // optimizes aggregates and choice rules. returns NULL if nothing optimizable.
    virtual varset* headVars() const = 0; // globally bound variables in head (if rule) or in weight@level (if weak constraint / optimize statement)
    virtual set<varset*>* elementVars() const = 0; // for each element (head & body elements) returns a varset of globally bound variables
    virtual set<varset*>* bodyVars() const = 0; // for each body element returns a varset of globally bound variables
    virtual void addToBody(NafLiteral *l) = 0; // adds the given literal to this statement's body.
    virtual vector<const BodyElement *>* selectBody(varset &s) const = 0; // returns those body elements whose global-bound variables are all included in s, but does not return elements that have no global-bound variables.
    virtual vector<BodyElement *>* selectBody(varset &s) = 0; // returns those body elements whose global-bound variables are all included in s, but does not return elements that have no global-bound variables.
    virtual vector<BodyElement *>* extractBody(varset &s) = 0; // like selectBody, but removes the returned elements from this.
    virtual vector<BodyElement *>* extractBodyWithoutGlobvarUpdate(varset &s) = 0; // like extractBody, but, for performance reasons, does not update this object's internal copy of global variables (if applicable). After this call, each time you use a function on this object, "globalVars" potentially means the old global vars (with potentially still more inside), until calcVariables is being called.
    virtual void calcVariables() = 0;
    virtual void write(ostream &stream) const = 0;
    virtual vector<const BodyElement *>* unsafe() const = 0; // returns all unsafe body elements
    virtual vector<const BodyElement *>* selectSafe(const Variable& v) const = 0; // returns a numer of body elements that make v safe. returns NULL if no such subset exists.  since multiple solutions may be possible, returns that one with the leas global variables.
    virtual vector<BodyElement *>* selectSafe(const Variable& v) = 0; // returns a numer of body elements that make v safe. returns NULL if no such subset exists.  since multiple solutions may be possible, returns that one with the leas global variables.
    virtual vector<const BodyElement *>* variableFreeBody() const = 0; // returns those body elements that have no global-bound variables.
    virtual Statement* cloneHead(Body* body) const = 0; // clones this statement, but replaces the body by the given.
  }; // abstract class
  
  class Program : public AspEntity
  {
    friend class parser::Parser;
  public:
    Program(vector<Statement *> *statements, vector<string *> *metastatements, ClassicalLiteral *query);
    Program(vector<Statement *> *statements, ClassicalLiteral *query);
    Program();
    Program(vector<Statement *> *statements, vector<string *> *metastatements);
    Program(vector<Statement *> *statements);
    Program(ClassicalLiteral *query);
    Program(Program const &that);
    virtual ~Program();
    Program& operator=(Program that);
    Program(Program&& other);
  public:
    friend void swap(Program& first, Program& second);
    void replaceStatements(vector<Statement *> *newStatements) const;
    void write(ostream &stream) const;
    Program* clone() const;
    void splitPooling(); // splits all pooling occurrences in statements into multiple statements. Afterwards, this program has no pooling anymore (apart from the ';' seperator for multiple elements within choices or aggregates)
    void optimizeInnerDomains(StringGenerator& gen); // optimizes aggregates and choice rules and inserts new rules for new temporary predicates. If nothing is optimizable, does nothing. Uses the given string generator for new predicates.
  public:
    const Statement& getStatement(int n) const;
    unsigned int statementsSize() const;
    vector<Statement *> getStatements();
    Program* cloneQueryAndMetastatements(vector<Statement*>* stmts) const; // clones this program except for its statements. The new program will have the statements given.
  private:
    vector<Statement *> *statements;
    vector<string *> *metastatements;
    ClassicalLiteral *query; // may be NULL
  };
  
  class WeightAtLevel : public AspEntity
  {
    friend class parser::Parser;
  public:
    WeightAtLevel(Term *weight, Term *level, TermTuple *tuple);
    WeightAtLevel(Term *weight);
    WeightAtLevel(Term *weight, Term *level);
    WeightAtLevel(Term *weight, TermTuple *tuple);
    WeightAtLevel(WeightAtLevel const &that);
    virtual ~WeightAtLevel();
    WeightAtLevel& operator=(WeightAtLevel that);
    WeightAtLevel(WeightAtLevel&& other);
  private:
    WeightAtLevel();
  public:
    friend void swap(WeightAtLevel& first, WeightAtLevel& second);
    void write(ostream &stream) const;
    varset* variables() const; // all variables occuring in here
    WeightAtLevel* clone() const;
    vector<WeightAtLevel*>* splitPooling();
  private:
    Term *weight;
    Term *level; // my be NULL
    TermTuple *tuple; //may be NULL
  };

  class LiteralBody : public AspEntity // for choice, aggregate and optimize elements
  {
    friend class parser::Parser;
  public:
    LiteralBody(vector<NafLiteral *> *literals);
    LiteralBody();
    LiteralBody(LiteralBody const &that);
    virtual ~LiteralBody();
    LiteralBody& operator=(LiteralBody that);
    LiteralBody(LiteralBody&& other);
    LiteralBody(Body* from); // takes the body elements of the given body as literals for this LiteralBody. then, from is destroyed.
  public:
    friend void swap(LiteralBody& first, LiteralBody& second);
    void write(ostream &stream) const;
    varset* variables() const; // all variables variables in all literals.
    LiteralBody* clone() const;
    vector<LiteralBody*>* splitPooling(); // normal behaviour (like most other classes)
    bool empty() const;
    SafetyCharacterization* safety(const varset& globalVars) const; // combined safety of all literals
    SafetyCharacterization* safety() const; // combined safety of all literals. Assumed that only this body defines globale variables.
    LiteralBody* splitContains(const varset& vars); // partitions this body into two bodies: those literals that contain at least one variable of vars stay in *this. The others, for a start, form the new literalBody that is returned. Literals in the new body that are unsafe w.r.t. the body context are pushed back into this.
    int size() const;
    Body* toRegularBody(); // constructs a regular body object that holds the same literal object as this. Afterwards, *this is in an inconsistent state and must be deleted.
    void insert(NafLiteral* literal); // inserts the literal at the end. Uses the given literal object. do not use it for anything else.
    set<varset*>* elementVars() const; // returns variables of each literal
    vector<const BodyElement *>* select(varset &s) const; // returns those elements whose global-bound variables are all included in s, but does not return elements that have no global-bound variables. Global in sense of this LiteralBody as body.
    vector<BodyElement *>* select(varset &s); // returns those elements whose global-bound variables are all included in s, but does not return elements that have no global-bound variables. Global in sense of this LiteralBody as body.
    vector<BodyElement *>* extract(varset &s); //like select, but removes returned literals from this.
    vector<BodyElement *>* extractWithoutGlobvarUpdate(varset &s); // like extract, for performance reasons, does not update this object's internal copy of global variables. After this call, each time you use a function on this object, "globalVars" potentially means the old global vars (with potentially still more inside), until calcVariables is being called.
    vector<const BodyElement *>* unsafe() const; // returns all unsafe literals
    vector<const BodyElement *>* variableFree() const; // returns those literals that have no global-bound variables.
    vector<const BodyElement *>* selectSafe(const Variable& v) const; // returns a numer of literals that make v safe. returns NULL if no such subset exists. since multiple solutions may be possible, returns that one with the leas global variables.
    vector<BodyElement *>* selectSafe(const Variable& v); // returns a numer of literals that make v safe. returns NULL if no such subset exists. since multiple solutions may be possible, returns that one with the leas global variables.
    void calcVariables(); // calculates the global variables of this body and stores them internally.
    void combine(LiteralBody* other); // moves the other's literals into this, deleting other.
  private:
    vector<NafLiteral *> *literals; // may be empty
    varset globalVariables;
  };

  class OptimizeElement : public AspEntity
  {
    friend class parser::Parser;
  public:
    OptimizeElement(WeightAtLevel *weightAtLevel, LiteralBody *literals);
    OptimizeElement(OptimizeElement const &that);
    virtual ~OptimizeElement();
    OptimizeElement& operator=(OptimizeElement that);
    OptimizeElement(OptimizeElement&& other);
  private:
    OptimizeElement();
  public:
    friend void swap(OptimizeElement& first, OptimizeElement& second);
    void write(ostream &stream) const;
    varset* variables() const; // weight@level ignored; union of all publically visible variables in the literals.
    OptimizeElement* clone() const;
    vector<OptimizeElement*>* splitPooling();
    SafetyCharacterization* safety() const; // safety of body
    varset* headVars() const; // returns vars in weight@level
    set<varset*>* elementVars() const; // for each element (w@l & body elements) returns a varset of those variables that also occur in variables()
    set<varset*>* bodyVars() const; // for each body element returns a varset of globally bound variables (i.e. those variables that also occur un variables())
    void addToBody(NafLiteral *l); // adds the given literal to the body.
    vector<const BodyElement *>* selectBody(varset &s) const; // returns those body elements whose global-bound variables are all included in s, but does not return elements that have no global-bound variables.
    vector<BodyElement *>* selectBody(varset &s); // returns those body elements whose global-bound variables are all included in s, but does not return elements that have no global-bound variables.
    vector<BodyElement *>* extractBody(varset &s);
    vector<const BodyElement *>* unsafe() const; // returns all unsafe body elements
    vector<const BodyElement *>* variableFreeBody() const; // returns those body elements that have no global-bound variables.
    OptimizeElement* cloneHead(Body* body) const; // clones this element, but replaces the body by the given.
    vector<const BodyElement *>* selectSafe(const Variable& v) const; // returns a numer of body elements that make v safe. returns NULL if no such subset exists.  since multiple solutions may be possible, returns that one with the leas global variables.
    vector<BodyElement *>* selectSafe(const Variable& v); // returns a numer of body elements that make v safe. returns NULL if no such subset exists.  since multiple solutions may be possible, returns that one with the leas global variables.
  private:
    WeightAtLevel *weightAtLevel;
    LiteralBody *literals;
  };
  
  class OptimizeStatement : public Statement
  {
    friend class parser::Parser;
  public:
    OptimizeStatement(string *optimizeType, vector<OptimizeElement *> *elements);
    OptimizeStatement(OptimizeStatement const &that);
    virtual ~OptimizeStatement();
    OptimizeStatement& operator=(OptimizeStatement that);
    OptimizeStatement(OptimizeStatement&& other);
  private:
    OptimizeStatement();
  public:
    friend void swap(OptimizeStatement& first, OptimizeStatement& second);
    void write(ostream &stream) const;
    varset* variables() const; // all publically visible variables in this statement - the union of the variables in the elements as defined there. Note that this makes no sense unless this statement has at most one element. Perform splitPooling() first to ensure this.
    OptimizeStatement* clone() const;
    vector<Statement*>* splitPooling(); // splits all optimize elements into their own optimize statement. afterwards, each optimize statement has max. 1 element.
    SafetyCharacterization* safety() const;
    vector<Rule*>* optimizeInnerDomains(StringGenerator& gen); // returns NULL because nothing optimizable.
    varset* headVars() const; // globally bound variables in weight@level
    set<varset*>* elementVars() const; // for each element (head & body elements) returns a varset of globally bound variables
    set<varset*>* bodyVars() const; // for each body element returns a varset of globally bound variables
    void addToBody(NafLiteral *l); // adds the given literal to this statement's body.
    vector<const BodyElement *>* selectBody(varset &s) const; // returns those body elements whose global-bound variables are all included in s, but does not return elements that have no global-bound variables.
    vector<BodyElement *>* selectBody(varset &s); // returns those body elements whose global-bound variables are all included in s, but does not return elements that have no global-bound variables.
    vector<BodyElement *>* extractBody(varset &s);
    vector<BodyElement *>* extractBodyWithoutGlobvarUpdate(varset &s); // like extractBody TODO may be optimized as with Body and LiteralBody
    void calcVariables(); // does nothing
    vector<const BodyElement *>* unsafe() const; // returns all unsafe body elements
    vector<const BodyElement *>* selectSafe(const Variable& v) const; // returns a numer of body elements that make v safe. returns NULL if no such subset exists. since multiple solutions may be possible, returns that one with the leas global variables.
    vector<BodyElement *>* selectSafe(const Variable& v); // returns a numer of body elements that make v safe. returns NULL if no such subset exists. since multiple solutions may be possible, returns that one with the leas global variables.
    vector<const BodyElement *>* variableFreeBody() const; // returns those body elements that have no global-bound variables.
    Statement* cloneHead(Body* body) const; // clones this statement, but replaces the body by the given.
  private:
    string *optimizeType;
    vector<OptimizeElement *> *elements;
  };
  
  class Body : public AspEntity
  {
    friend class parser::Parser;
    friend class LiteralBody;
  public:
    Body();
    Body(vector<BodyElement *> *elements);
    Body(Body const &that);
    virtual ~Body();
    Body& operator=(Body that);
    Body(Body&& other);
  public:
    friend void swap(Body& first, Body& second);
    void write(ostream &stream) const;
    varset* variables() const; // all global variables in this body.
    Body* clone() const;
    bool empty() const;
    vector<Body*>* splitPooling();
    SafetyCharacterization* safety() const; // combined safety of all body elements
    SafetyCharacterization* safety(const varset &globalVars) const; // like safety(), but considering the given variables as the global variables in this body instead of the actual global variables
    vector<Rule*>* optimizeInnerDomains(StringGenerator& gen); // for each element, optimizes inner domains and combines the results. If nothing can be optimized, returns NULL.
    void add(BodyElement *el); // appends the element
    set<varset*>* elementVars() const; // for each element returns a varset of globally bound variables
    vector<const BodyElement *>* select(varset &s) const; // returns those elements whose global-bound variables are all included in s, but does not return elements that have no global-bound variables.
    vector<BodyElement *>* select(varset &s); // returns those elements whose global-bound variables are all included in s, but does not return elements that have no global-bound variables.
    vector<BodyElement *>* extract(varset &s); // like select, but removes returned elements from this.
    vector<BodyElement *>* extractWithoutGlobvarUpdate(varset &s); // like extract, for performance reasons, does not update this object's internal copy of global variables. After this call, each time you use a function on this object, "globalVars" potentially means the old global vars (with potentially still more inside), until calcVariables is being called.
    vector<const BodyElement *>* unsafe() const; // returns all unsafe body elements
    vector<const BodyElement *>* selectSafe(const Variable& v) const; // returns a numer of body elements that make v safe. returns NULL if no such subset exists. since multiple solutions may be possible, returns that one with the least global variables.
    vector<BodyElement *>* selectSafe(const Variable& v); // returns a numer of body elements that make v safe. returns NULL if no such subset exists. since multiple solutions may be possible, returns that one with the least global variables.
    vector<const BodyElement *>* variableFree() const; // returns those elements that have no global-bound variables.
    void calcVariables(); // calculates the global variables of this body and stores them internally.
  private:
    vector<BodyElement *> *elements;
    varset globalVariables;
  };
  
  class WeakConstraint : public Statement
  {
    friend class parser::Parser;
  public:
    WeakConstraint(Body *body, WeightAtLevel *weightAtLevel);
    WeakConstraint(WeakConstraint const &that);
    virtual ~WeakConstraint();
    WeakConstraint& operator=(WeakConstraint that);
    WeakConstraint(WeakConstraint&& other);
  private:
    WeakConstraint();
  public:
    friend void swap(WeakConstraint& first, WeakConstraint& second);
    void write(ostream &stream) const;
    varset* variables() const; // all variables in this statement's body, ignoring weight@level.
    WeakConstraint* clone() const;
    vector<Statement*>* splitPooling();
    SafetyCharacterization* safety() const;
    vector<Rule*>* optimizeInnerDomains(StringGenerator& gen); // optimizes aggregates. If nothing is optimizable, returns NULL.
    varset* headVars() const; // globally bound variables in head (if rule) or in weight@level (if weak constraint / optimize statement)
    set<varset*>* elementVars() const; // for each element (head & body elements) returns a varset of globally bound variables
    set<varset*>* bodyVars() const; // for each body element returns a varset of globally bound variables
    void addToBody(NafLiteral *l); // adds the given literal to this statement's body.
    vector<const BodyElement *>* selectBody(varset &s) const; // returns those body elements whose global-bound variables are all included in s, but does not return elements that have no global-bound variables.
    vector<BodyElement *>* selectBody(varset &s); // returns those body elements whose global-bound variables are all included in s, but does not return elements that have no global-bound variables.
    vector<BodyElement *>* extractBody(varset &s);
    vector<BodyElement *>* extractBodyWithoutGlobvarUpdate(varset &s); // like extractBody TODO may be optimized as with Body and LiteralBody
    void calcVariables(); // does nothing
    vector<const BodyElement *>* unsafe() const; // returns all unsafe body elements
    vector<const BodyElement *>* selectSafe(const Variable& v) const; // returns a numer of body elements that make v safe. returns NULL if no such subset exists. since multiple solutions may be possible, returns that one with the leas global variables.
    vector<BodyElement *>* selectSafe(const Variable& v); // returns a numer of body elements that make v safe. returns NULL if no such subset exists. since multiple solutions may be possible, returns that one with the leas global variables.
    vector<const BodyElement *>* variableFreeBody() const; // returns those body elements that have no global-bound variables.
    Statement* cloneHead(Body* body) const; // clones this statement, but replaces the body by the given.
  private:
    Body *body;
    WeightAtLevel *weightAtLevel;
  };
  
  class RuleHead : public AspEntity
  {
    friend class parser::Parser;
  public:
    virtual ~RuleHead();
    virtual varset* variables() const = 0; // all variables in this head
    virtual varset* variables(const varset& globalVars) const; // all publically visible variables in this head
    virtual RuleHead* clone() const = 0; // deep copy
    virtual vector<RuleHead*>* splitPooling() = 0; // splits this head into multiple heads if it contains pooling.
    virtual vector<Rule*>* optimizeInnerDomains(StringGenerator& gen, const varset& globalVars) = 0; // optimizes choice rules. If nothing is optimizable, returns NULL.
    virtual bool empty() = 0; // if when writing, nothing is to be written.
  }; // abstract class

  class Rule : public Statement
  {
    friend class parser::Parser;
  public:
    Rule(RuleHead *head, Body *body);
    Rule(Rule const &that);
    virtual ~Rule();
    Rule& operator=(Rule that);
    Rule(Rule&& other);
  private:
    Rule();
  public:
    friend void swap(Rule& first, Rule& second);
    void write(ostream &stream) const;
    varset* variables() const; // all public variables in the body only; the head is ignored.
    Rule* clone() const;
    vector<Statement*>* splitPooling();
    SafetyCharacterization* safety() const;
    vector<Rule*>* optimizeInnerDomains(StringGenerator& gen); // optimizes aggregates and choice rules. If nothing is optimizable, returns NULL.
    varset* headVars() const; // globally bound variables in head (if rule) or in weight@level (if weak constraint / optimize statement)
    set<varset*>* elementVars() const; // for each element (head & body elements) returns a varset of globally bound variables
    set<varset*>* bodyVars() const; // for each body element returns a varset of globally bound variables
    void addToBody(NafLiteral *l); // adds the given literal to this statement's body.
    vector<const BodyElement *>* selectBody(varset &s) const; // returns those body elements whose global-bound variables are all included in s, but does not return elements that have no global-bound variables.
    vector<BodyElement *>* selectBody(varset &s); // returns those body elements whose global-bound variables are all included in s, but does not return elements that have no global-bound variables.
    vector<BodyElement *>* extractBody(varset &s); // see Statement
    vector<BodyElement *>* extractBodyWithoutGlobvarUpdate(varset &s); // see Statement
    void calcVariables(); // see Statement
    vector<const BodyElement *>* unsafe() const; // returns all unsafe body elements
    vector<const BodyElement *>* selectSafe(const Variable& v) const; // returns a numer of body elements that make v safe. returns NULL if no such subset exists. since multiple solutions may be possible, returns that one with the leas global variables.
    vector<BodyElement *>* selectSafe(const Variable& v); // returns a numer of body elements that make v safe. returns NULL if no such subset exists. since multiple solutions may be possible, returns that one with the leas global variables.
    vector<const BodyElement *>* variableFreeBody() const; // returns those body elements that have no global-bound variables.
    Statement* cloneHead(Body* body) const; // clones this statement, but replaces the body by the given.
  private:
    RuleHead *head;
    Body *body;
  };
  
  class Disjunction : public RuleHead
  {
    friend class parser::Parser;
  public:
    Disjunction(vector<ClassicalLiteral *> *literals);
    Disjunction(Disjunction const &that);
    virtual ~Disjunction();
    Disjunction& operator=(Disjunction that);
    Disjunction(Disjunction&& other);
  private:
    Disjunction();
  public:
    friend void swap(Disjunction& first, Disjunction& second);
    void write(ostream &stream) const;
    varset* variables() const; // all variables in here.
    Disjunction* clone() const;
    vector<RuleHead*>* splitPooling();
    vector<Rule*>* optimizeInnerDomains(StringGenerator& gen, const varset& globalVars); // nothing is optimizable, returns NULL.
    bool empty(); // if when writing, nothing is to be written.
  private:
    vector<ClassicalLiteral *> *literals; // may be empty!
  };

  class Guard : public AspEntity
  {
    friend class parser::Parser;
  public:
    Guard(string *binop, Term *term, bool isLeft);
    Guard(Guard const &that);
    virtual ~Guard();
    Guard& operator=(Guard that);
    Guard(Guard&& other);
  private:
    Guard();
  public:
    friend void swap(Guard& first, Guard& second);
    bool isLeft() const;
    void write(ostream &stream) const;
    varset* variables() const;
    Guard* clone() const;
    bool equal() const;
    vector<Guard*>* splitPooling();
    SafetyCharacterization* safety() const; // returns a characterization of the term. No safety rules included.
  private:
    string *binop;
    Term *term;
    bool left;
  };

  class ChoiceElement : public AspEntity
  {
    friend class parser::Parser;
  public:
    ChoiceElement(ClassicalLiteral *literal, LiteralBody *dependencyLiterals);
    ChoiceElement(ChoiceElement const &that);
    virtual ~ChoiceElement();
    ChoiceElement& operator=(ChoiceElement that);
    ChoiceElement(ChoiceElement&& other);
  private:
    ChoiceElement();
  public:
    friend void swap(ChoiceElement& first, ChoiceElement& second);
    void write(ostream &stream) const;
    varset* variables() const;
    ChoiceElement* clone() const;
    vector<ChoiceElement*>* splitPooling(); // splits as usual.
    Rule* optimizeInnerDomain(StringGenerator& gen, const varset& globalVars); // if this is not an optimizable element, does nothing (returns NULL). If optimizable, changes it (introduces a new predicate) and returns the induction rule for the new predicate.
  private:
    ClassicalLiteral *literal;
    LiteralBody *dependencyLiterals;
  };

  class Choice : public RuleHead
  {
    friend class parser::Parser;
  public:
    Choice(vector<ChoiceElement *> *elements, Guard *guard_l, Guard *guard_r); // g_l one must be left, g_r must be right guard
    Choice(vector<ChoiceElement *> *elements);
    Choice(vector<ChoiceElement *> *elements, Guard *guard);
    Choice(Choice const &that);
    virtual ~Choice();
    Choice& operator=(Choice that);
    Choice(Choice&& other);
  private:
    Choice();
  public:
    friend void swap(Choice& first, Choice& second);
    void write(ostream &stream) const;
    varset* variables() const;
    Choice* clone() const;
    vector<RuleHead*>* splitPooling(); // choice element alternatives are split inside the choice rule, not leading to more RuleHeads. Only pooling in the guards may lead to more rule heads.
    vector<Rule*>* optimizeInnerDomains(StringGenerator& gen, const varset& globalVars); // optimizes this choice rule. If nothing is optimizable, returns NULL.
    bool empty(); // if when writing, nothing is to be written.
  private:
    vector<ChoiceElement *> *elements;
    Guard *guard_l, *guard_r; // each may be null
  };

  class AggregateElement : public AspEntity
  {
    friend class parser::Parser;
  public:
    AggregateElement(TermTuple *termTuple, LiteralBody *dependencyLiterals);
    AggregateElement(TermTuple *termTuple);
    AggregateElement(LiteralBody *dependencyLiterals);
    AggregateElement(AggregateElement const &that);
    virtual ~AggregateElement();
    AggregateElement& operator=(AggregateElement that);
    AggregateElement(AggregateElement&& other);
  private:
    AggregateElement();
  public:
    friend void swap(AggregateElement& first, AggregateElement& second);
    void write(ostream &stream) const;
    varset* variables() const; // all variables in this element
    AggregateElement* clone() const;
    vector<AggregateElement*>* splitPooling(); // normal behaviour (like most other classes)
    Rule* optimizeInnerDomain(StringGenerator& gen, const varset& globalVars); // if this is not an optimizable aggregate element, does nothing (returns NULL). If optimizable, changes it (introduces a new predicate) and returns the induction rule for the new predicate.
  private:
    TermTuple *termTuple; // may be NULL
    LiteralBody *dependencyLiterals;
  };

  class Aggregate : public BodyElement
  {
    friend class parser::Parser;
  public:
    Aggregate(string *function, bool nafNegated, vector<AggregateElement *> *elements, Guard *guard_l, Guard *guard_r); // function has to include '#' (like "#sum"). g_l one must be left, g_r must be right guard
    Aggregate(string *function, bool nafNegated, vector<AggregateElement *> *elements, Guard *guard); // function has to include '#' (like "#sum")
    Aggregate(string *function, bool nafNegated, vector<AggregateElement *> *elements); // function has to include '#' (like "#sum")
    Aggregate(Aggregate const &that);
    virtual ~Aggregate();
    Aggregate& operator=(Aggregate that);
    Aggregate(Aggregate&& other);
  private:
    Aggregate();
  public:
    friend void swap(Aggregate& first, Aggregate& second);
    void write(ostream &stream) const;
    varset* variables(const varset& globalVars) const; // all publically visible variables in this body element
    virtual bool globalVarsNotEmptyAndIncludedIn(const varset& globalVariables, const varset& s) const; // checks if the variables of variables(globalVars) are not empty and fully included in s
    Aggregate* clone() const;
    vector<BodyElement*>* splitPooling(); // splits all poolings inside the aggregate by introducing new aggregate elements into this. then, the guards are considered. If a guard has pooling, the returning vector will be non-empty.
    vector<Rule*>* optimizeInnerDomains(StringGenerator& gen, const varset& globalVars); // if this is not an optimizable aggregate, does nothing (returns NULL). If an optimizable aggregate, changes it (introduces new predicates) and returns the induction rules for the new predicates.
    const SafetyCharacterization* safety(const varset& globalVars); // all globally bound variables inside the aggregate elements are unsafe. guards: same as BuiltinAtom.
    void dispose(const SafetyCharacterization* safety) const;
  private:
    varset* insideVariables() const; // all variables inside the elements
    void updatePublicVariables(); // sets publicVariables to those of the guards
  private:
    string *function;
    bool nafNegated;
    vector<AggregateElement *> *elements;
    Guard *guard_l, *guard_r; // each may be null
  };

  class BuiltinAtom: public NafLiteral
  {
    friend class parser::Parser;
  public:
    BuiltinAtom(Term *t1, string *binop, Term *t2);
    BuiltinAtom(BuiltinAtom const &that);
    virtual ~BuiltinAtom();
    BuiltinAtom& operator=(BuiltinAtom  that);
    BuiltinAtom(BuiltinAtom&& other);
  private:
    BuiltinAtom();
  public:
    friend void swap(BuiltinAtom& first, BuiltinAtom& second);
    void write(ostream &stream) const;
    BuiltinAtom* clone() const;
    bool equal() const;
    template <typename T> vector<T *> *splitPooling(); // T must be supertype of BuiltinAtom
    vector<BodyElement*>* splitPooling();
    vector<NafLiteral*>* splitPoolingNaf();
  protected:
    virtual const SafetyCharacterization* computeSafetyCharacterization() const;
  private:
    void updatePublicVariables(); // sets publicVariables
  private:
    Term *t1, *t2;
    string *binop;
  };  
  


} // namespace domain

#endif // ASP_DOMAIN_H
