/*
Copyright 2015-2016, Manuel Bichler
WWW: <http://dbai.tuwien.ac.at/proj/lpopt/>

This file is part of lpopt.

lpopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

lpopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lpopt.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <fstream>
#include <algorithm>
#include <list>

#include "LogicProgramOptimizationManager.hpp"

using namespace std;
using namespace lpopt;

LogicProgramOptimizationManager::LogicProgramOptimizationManager(std::istream *stream, DecompositionOrderingAlgorithm orderingAlg, bool ignoreHeadVariables) : ignoreHeadVariables(ignoreHeadVariables), parsingDriver(stream), orderingAlg(orderingAlg), program(NULL), width(0)
{
}

LogicProgramOptimizationManager::LogicProgramOptimizationManager(std::istream *stream, DecompositionOrderingAlgorithm orderingAlg, bool ignoreHeadVariables, std::istream *instStream, char* macroFileName, std::string keyListIn) : ignoreHeadVariables(ignoreHeadVariables), parsingDriver(stream), orderingAlg(orderingAlg), program(NULL), width(0)
{	
	char c;
	ostringstream tempStream;
	while(instStream->get(c))
	{
		tempStream << c;
	}
	instanceString = tempStream.str();
	tempStream.flush();
	macroFile = macroFileName;
	keyList = keyListIn;
	
}

LogicProgramOptimizationManager::~LogicProgramOptimizationManager()
{
	delete program;
}

//#pragma GCC diagnostic ignored "-Wmaybe-uninitialized" // beg
Program* LogicProgramOptimizationManager::optimize()
{
	checkParsedAndPreprocessed();

//	clock_t beg;
//	if (benchmark) beg = clock();

	unique_ptr<htd::LibraryInstance> htd(htd::createManagementInstance(htd::Id::FIRST));
	htd->treeDecompositionAlgorithmFactory().setConstructionTemplate(new BucketEliminationTreeDecompositionAlgorithm(htd.get()));
	IOrderingAlgorithm *orderAlg;
	switch(orderingAlg)
	{
        case MCS:
		orderAlg = new MaximumCardinalitySearchOrderingAlgorithm(htd.get());
                break;
        case MF:
		orderAlg = new MinFillOrderingAlgorithm(htd.get());
                break;
	default:
        case MIW:
		orderAlg = new MinDegreeOrderingAlgorithm(htd.get());
		break;
        case NOA:
		orderAlg = new NaturalOrderingAlgorithm(htd.get());
                break;
        }
	htd->orderingAlgorithmFactory().setConstructionTemplate(orderAlg);

	StatementOptimizationAlgorithm optAlg;

	list<Statement *> createdStatements;

	for (unsigned int curStmt = 0; curStmt < program->statementsSize(); ++curStmt)
	{ // decompose each statement and put results in createdStatements
		string* genString = new string("temp_decomp_");
		genString->append(std::to_string(curStmt)).append("_");
		StringGenerator gen(genString);

		HypergraphableStatement st(program->getStatement(curStmt), ignoreHeadVariables, gen);

/*		if(benchmark)
		{
			me_time += clock() - beg;
			beg = clock();
		}*/

		unique_ptr<IHypergraph> hg(st.toHypergraph(*htd));

		unique_ptr<htd::ITreeDecompositionAlgorithm> treeDecompositionAlgorithm(htd->treeDecompositionAlgorithmFactory().getTreeDecompositionAlgorithm());
		treeDecompositionAlgorithm->addManipulationOperation(st.createHeadRootingOperation(htd.get()));
		unique_ptr<ITreeDecomposition> decomp(treeDecompositionAlgorithm->computeDecomposition(*hg));
		if (decomp->maximumBagSize() > width) width = decomp->maximumBagSize();

/*		if(benchmark)
		{
			htdlib_time += clock() - beg;
			beg = clock();
		}*/


		unique_ptr<list<Statement *>> sol(optAlg.optimize(*decomp, st));

		//begin predictor manipulation
		
		if(!instanceString.empty())
		{
			Statement* currentStatement = program->getStatement(curStmt).clone();

			ostringstream regularStream;
			currentStatement->write(regularStream);

			ostringstream decompStream;
		
			for(unsigned int a = 0; a < sol->size(); a ++)
   			{
    			std::list<Statement *>::iterator i = sol->begin();

    			advance(i, a);
    			Statement* totry = *i;
    			totry->write(decompStream);
    		
    		}


			if(decompStream.str().compare(regularStream.str()) != 0)
			{	
				decompStream.flush();
				regularStream.flush();

				program->write(regularStream);

				vector<Statement *> curStatements = program->clone()->getStatements();
				
				int loopcount = 0;
				for(unsigned int a = 0; a < sol->size(); a++)
				{
					std::list<Statement *>::iterator i = sol->begin();

					advance(i, a);
					Statement* iterStatement = *i;
					if(a == 0)
					{
						curStatements.at(curStmt) = iterStatement;
					}
					else
					{
						curStatements.insert(curStatements.begin() + curStmt + a, iterStatement);
					}
					//cout << decompStream.str() << endl;
					loopcount = a;
				
				}
				

				Program* toPrint = program->clone()->cloneQueryAndMetastatements(&curStatements);

				toPrint->write(decompStream);

				// std::cout << "REGULAR~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << std::flush;
				// std::cout << regularStream.str() << std::flush;
				// std::cout << "DECOMP~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << std::flush;
				// std::cout << decompStream.str() << std::flush;

				// std::cout << "Begin predict on decomp \n" << std::flush;
				long decompValue;
				if (curStmt > 0){
					decompValue = Predictor.predict(decompStream.str(), instanceString, macroFile, keyList);
				}
				else{
					decompValue = Predictor.predict(decompStream.str(), instanceString, NULL, keyList);
				}
				decompStream.flush();
				// std::cout << "Finished! predict on decomp \n" << std::flush;

				// std::cout << "Begin predict on regular \n" << std::flush;
				long regularValue = Predictor.predict(regularStream.str(), instanceString, NULL, keyList);
				regularStream.flush();
				// std::cout << "Finished! predict on regular \n" << std::flush;

				Predictor.reset();
				
				if(decompValue <= regularValue)
				{	

					program->replaceStatements(&curStatements);
					curStmt += loopcount;

					//for (int i = 0; i < program->getStatements().size(); i++)
					//{
					//	program->getStatements().at(i)->write(cout);
					//	cout << "next" << endl;
					//}
				}

			}

			decompStream.flush();
			regularStream.flush();

		}
		else
		{
			createdStatements.splice(createdStatements.end(), *sol);
		}

		//cout << "after pass number: " << curStmt << endl;
		//program->write(cout);
		//cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl << endl << endl;

	}
	
	if(!instanceString.empty()) 
	{
		Predictor.~predictor();
	}
	else
	{
		vector<Statement *>* newStmtVector = new vector<Statement *>(createdStatements.begin(), createdStatements.end());

		Program* toReturn = program->cloneQueryAndMetastatements(newStmtVector);

		toReturn->write(cout);

		return toReturn;
	}
	

/*	if (benchmark)
	{
	  me_time += clock() - beg;
	  cout << "done optimizing logic program. took " << (float) (htdlib_time + me_time) / CLOCKS_PER_SEC << " seconds." << endl;
	  cout << "\t" << (float) me_time / CLOCKS_PER_SEC << " seconds for inner domain optimization, pooling splitting and final solution calculation based on decomp," << endl;
	  cout << "\t" << (float) htdlib_time / CLOCKS_PER_SEC << " seconds for hypergraph representation building and tree decomposing library." << endl;
	}*/

	cout << std::flush;

	program->write(cout);

	return program;
}

const Program& LogicProgramOptimizationManager::getBaseProgram()
{
	checkParsedAndPreprocessed();
	return *program;
}

void LogicProgramOptimizationManager::checkParsedAndPreprocessed()
{
	if (!program)
	{
		parse();
		preprocess();
	}
}

//#pragma GCC diagnostic ignored "-Wmaybe-uninitialized" // beg
void LogicProgramOptimizationManager::parse()
{
//	clock_t beg;
//	if (benchmark) beg = clock();

	delete program;
	program = parsingDriver.parse();	

/*	if (benchmark)
	{
	  clock_t thisTime = clock() - beg;
	  parse_time += thisTime;
	  cout << "parsing finished. took " << (float) thisTime / CLOCKS_PER_SEC << " seconds." << endl;
	}*/
}

//#pragma GCC diagnostic ignored "-Wmaybe-uninitialized" // beg
void LogicProgramOptimizationManager::preprocess()
{
//	clock_t beg;
//	if (benchmark) beg = clock();

	program->splitPooling();
	StringGenerator innerGen(new string("temp_innerDomain_"));
	program->optimizeInnerDomains(innerGen);

/*	if (benchmark)
	{
	  clock_t thisTime = clock() - beg;
	  me_time += thisTime;
	  cout << "\t\t" << (float) thisTime / CLOCKS_PER_SEC << " seconds for inner domain optimization and pooling splitting. (finished)" << endl;
	}*/
}

unsigned int LogicProgramOptimizationManager::getMaxWidth() const
{
	return width;
}
