#include <python3.6/Python.h>
#include <iostream>
#include <algorithm>
#include "predictor.hpp"
#include "frameobject.h"


using namespace std;


predictor::predictor()
{
    Py_Initialize();

    pName_networkx = PyUnicode_FromString("networkx");

    pModule_networkx = PyImport_Import(pName_networkx);

    Py_XDECREF(pName_networkx);

    pName = PyUnicode_FromString("predictor");

    pModule = PyImport_Import(pName);

    Py_XDECREF(pName);

    if (pModule == nullptr) {
    PyErr_Print();
    std::cerr << "Fails to import the module predictor, check installation.\n";
    }

    // dict is a borrowed reference.
    dict = PyModule_GetDict(pModule);
    if (dict == nullptr) {
        PyErr_Print();
        std::cerr << "Fails to get the dictionary, check predictor installation.\n";
        
        Py_XDECREF(pModule_networkx);
        Py_XDECREF(pModule);

        return;
    }
    

    // Builds the name of a callable class && a borrowed reference
    python_class = PyDict_GetItemString(dict, "Predictor");
    if (python_class == nullptr || python_class == NULL) {
        PyErr_Print();
        std::cerr << "Fails to get the Python class, check predictor installation.\n";
        Py_XDECREF(pModule_networkx);
        Py_XDECREF(pModule);
        return;
    }
    

    // Creates an instance of the class && a new reference
    if (PyCallable_Check(python_class)) {
        object = PyObject_CallObject(python_class, nullptr);
        if (object == NULL)
        {
            Py_XDECREF(python_class);
            cerr << "Fails to create object.";
            Py_XDECREF(pModule);
            Py_XDECREF(pModule_networkx);
            return;
        }
        
    } else {
        PyErr_Print();
        std::cout << "Cannot instantiate the Python class" << std::endl;
        Py_XDECREF(pModule);
        Py_XDECREF(pModule_networkx);
        return;
    }

    pMethod_predict = PyUnicode_FromString("predict_loaded_program");
    pMethod_load = PyUnicode_FromString("load_program_string");
    pMethod_load_file = PyUnicode_FromString("load_program");
    pMethod_reset = PyUnicode_FromString("reset_program");
    pMethod_reset_all = PyUnicode_FromString("reset_all");
    pMethod_set_keys = PyUnicode_FromString("set_keys");
}

void predictor::reset(){
    PyObject_CallMethodObjArgs(object, pMethod_reset_all, NULL);
}

predictor::~predictor()
{   

    Py_FinalizeEx();

}

long predictor::predict(string program, string instance, char* macroFile = NULL, std::string keyList)
{
    
    //std::cout << "Instance: " << instance.c_str() << '\n';
    
    if(keyList != ""){
        
        // std::cout << "C KEY LIST:    " << keyList << '\n' << std::flush;
        PyObject *pKeyList = PyUnicode_FromString(keyList.c_str());
        PyObject *keyList_load = PyObject_CallMethodObjArgs(object, pMethod_set_keys, pKeyList, NULL);

        if(keyList_load == NULL){
            printf("Error on keyList load\n");
            PyErr_Print();
            Py_XDECREF(pKeyList);
            Py_XDECREF(keyList_load);
            return -1;
        }
    }

    PyObject *pProgram = PyUnicode_FromString(program.c_str());
    PyObject *pInstance = PyUnicode_FromString(instance.c_str());
    // std::cout << "Load Program \n" << std::flush;
    PyObject *program_load = PyObject_CallMethodObjArgs(object, pMethod_load, pProgram, NULL);
    if(program_load == NULL){
        printf("Error on program load\n");
        PyErr_Print();
        Py_XDECREF(pProgram); 
        Py_XDECREF(pInstance); 
        Py_XDECREF(program_load); 
        return -1;
    }
    // std::cout << "Loaded Program \n" << std::flush;
    // std::cout << "Load Instance \n" << std::flush;
    PyObject *instance_load = PyObject_CallMethodObjArgs(object, pMethod_load, pInstance, NULL);

    if(instance_load == NULL){
        printf("Error on instance load");
        PyErr_Print();
        Py_XDECREF(pProgram);
        Py_XDECREF(pInstance); 
        Py_XDECREF(instance_load);
        Py_XDECREF(program_load); 
        return -1;
    }
    // std::cout << "Loaded Instance \n" << std::flush;

    if(macroFile != NULL){
        // std::cout << "Load Macro \n" << std::flush;
        PyObject *pMacro = PyUnicode_FromString(macroFile);
        PyObject *MacroFile_load = PyObject_CallMethodObjArgs(object, pMethod_load_file, pMacro, NULL);

        if(MacroFile_load == NULL){
            printf("Error on Macro Load");
            PyErr_Print();
            Py_XDECREF(pProgram);
            Py_XDECREF(pInstance);
            Py_XDECREF(pMacro); 
            Py_XDECREF(instance_load);
            Py_XDECREF(program_load); 
            Py_XDECREF(MacroFile_load); 
        return -1;
        }
        Py_XDECREF(pMacro);
        Py_XDECREF(MacroFile_load);
        // std::cout << "Loaded Macro \n" << std::flush;
    }


    // std::cout << "Begin Prediction! \n" << std::flush;
    PyObject *value = PyObject_CallMethodObjArgs(object, pMethod_predict, NULL);
    PyObject_CallMethodObjArgs(object, pMethod_reset, NULL);

    long int endValue = PyLong_AsLong(value);
    // std::cout << "finished Prediction! \n" << std::flush;

    // printf("Value: %ld \n", endValue);
    // std::cout << std::flush;

    if (endValue == -1)
    {
        if(!PyErr_Occurred())
        {

            PyErr_Print();
            cerr << "end Value < 0 and no pyError occured";

            Py_XDECREF(pProgram);
            Py_XDECREF(pInstance);
            Py_XDECREF(instance_load);
            Py_XDECREF(program_load); 
            
            return -1;
        }
        
        printf("Error on predictor call \n");

        PyObject *type, *ivalue, *traceback;
        PyErr_Fetch(&type, &ivalue, &traceback);

        const char* s = PyUnicode_AsUTF8(ivalue);
        std::cout << "Error: " << s << '\n' << std::flush;

        const char* t = PyUnicode_AsUTF8(type);
        std::cout << "Type: " << t << '\n' << std::flush;


        Py_XDECREF(type);
        Py_XDECREF(ivalue);
        Py_XDECREF(traceback);

        
    }

    Py_XDECREF(value);
    Py_XDECREF(pProgram); 
    Py_XDECREF(pInstance);
    Py_XDECREF(instance_load);
    Py_XDECREF(program_load);

    return endValue;
}